use std::collections::{HashMap, VecDeque};

use based::{get_pg, request::api::ToAPI, result::LogNoneAndPass, ui::components::prelude::Shell};
use ollama_rs::generation::embeddings::request::{EmbeddingsInput, GenerateEmbeddingsRequest};
use serde::Serialize;
use serde_json::json;
use sqlx::FromRow;

use crate::{
    archive::{Document, Domain, WebsiteArchive},
    conf::get_config,
};

// TODO : Cite found chunks in search res?

#[derive(Debug, Clone, FromRow, Serialize)]
pub struct DocEmbedding {
    pub domain: String,
    pub path: String,
    pub ver: String,
    pub chunk: i32,

    #[allow(dead_code)]
    #[serde(skip)]
    embed_mxbai_embed_large: pgvector::Vector,

    #[sqlx(default)]
    pub similarity: f64,
}

impl DocEmbedding {
    pub async fn total_chunks(&self) -> i64 {
        let res: (i64,) = sqlx::query_as(
            "SELECT COUNT(chunk) FROM doc_embedding WHERE domain = $1 AND path = $2",
        )
        .bind(&self.domain)
        .bind(&self.path)
        .fetch_one(get_pg!())
        .await
        .unwrap();

        res.0
    }
}

#[derive(Debug, Clone, Serialize)]
pub struct SearchResult {
    pub domain: String,
    pub path: String,
    pub total_chunks: i64,
    pub chunks: Vec<DocEmbedding>,
}

impl SearchResult {
    pub fn new(domain: String, path: String, total_chunks: i64) -> Self {
        Self {
            domain,
            path,
            total_chunks,
            chunks: vec![],
        }
    }

    pub fn similarity(&self) -> f64 {
        let chunks = f64::from(self.chunks.len() as u32);
        let total = f64::from(self.total_chunks as i32);
        let match_percent = chunks / total;
        total_score(&self.chunks) * match_percent
    }
}

pub fn avg_sim(e: &[DocEmbedding]) -> f64 {
    let mut score = 0.0;

    for e in e {
        score += e.similarity;
    }

    score / e.len() as f64
}

pub fn max_sim(e: &[DocEmbedding]) -> f64 {
    let mut score = 0.0;

    for e in e {
        if e.similarity > score {
            score = e.similarity;
        }
    }

    score
}

pub fn total_score(e: &[DocEmbedding]) -> f64 {
    (avg_sim(e) + max_sim(e)) / 2.0
}

impl ToAPI for DocEmbedding {
    async fn api(&self) -> serde_json::Value {
        json!({
            "domain": self.domain,
            "path": self.path,
            "ver": self.ver,
            "similarity": self.similarity
        })
    }
}

pub trait Embedding {
    fn embedding(
        &self,
        ver: Option<String>,
        shell: &Shell,
    ) -> impl std::future::Future<Output = Option<Vec<Vec<f32>>>>;
}

pub fn chunked(s: &str) -> Vec<String> {
    const CHUNK_SIZE: usize = 2500;
    s.chars()
        .collect::<Vec<char>>()
        .chunks(CHUNK_SIZE)
        .map(|chunk| chunk.iter().collect())
        .collect()
}

pub fn remove_data_urls(input: &str) -> String {
    let re = regex::Regex::new("data:(.*?)(;base64)?,(.*)").unwrap();

    // Replace all occurrences of data URLs with an empty string
    re.replace_all(input, "").to_string()
}

impl Embedding for Document {
    async fn embedding(&self, ver: Option<String>, shell: &Shell) -> Option<Vec<Vec<f32>>> {
        let latest = "latest".to_string();
        log::info!(
            "Generating Vector embeddings for {} / {} @ {}",
            self.domain,
            self.path,
            ver.as_ref().unwrap_or(&latest)
        );

        let content_html = self.render_local(ver.clone(), shell).await.ok()?;
        let content_html = String::from_utf8_lossy(&content_html);
        let content = remove_data_urls(&html2md::parse_html(&content_html));

        let mut embeddings = Vec::new();
        let content = chunked(&content);
        let len = content.len();

        for (index, c) in content.into_iter().enumerate() {
            log::info!(
                "Generating Vector embeddings for {} / {} @ {} [ {} / {} ]",
                self.domain,
                self.path,
                ver.as_ref().unwrap_or(&latest),
                index + 1,
                len
            );
            embeddings.push(generate_embedding(c).await?);
        }

        Some(embeddings)
    }
}

pub async fn generate_embedding(mut input: String) -> Option<Vec<f32>> {
    // TODO : Ollama load balancing
    if let Some(ollama_url) = get_config().ai.as_ref().map(|x| x.OLLAMA_URL.clone()) {
        let (host, port) = ollama_url.split_once(':')?;
        let ollama = ollama_rs::Ollama::new(format!("http://{host}"), port.parse().ok()?);

        let models = ollama.list_local_models().await.ok()?;

        if !models
            .into_iter()
            .any(|x| x.name.starts_with("mxbai-embed-large"))
        {
            log::info!("Model not found. Pulling 'mxbai-embed-large'");
            ollama
                .pull_model("mxbai-embed-large".to_string(), false)
                .await
                .ok()?;
        }

        if input.is_empty() {
            input = " ".to_string();
        }

        let res = ollama
            .generate_embeddings(GenerateEmbeddingsRequest::new(
                "mxbai-embed-large".to_string(),
                EmbeddingsInput::Single(input),
            ))
            .await
            .ok()?;
        let embed = res.embeddings.first()?;
        return Some(embed.clone());
    }

    None
}

pub struct EmbedStore;

impl EmbedStore {
    pub async fn get_embedding(doc: &Document, ver: Option<&str>) -> Option<DocEmbedding> {
        let use_ver = ver.map_or_else(
            || {
                let version = doc.versions();
                version.first().unwrap().clone()
            },
            |x| x.to_string(),
        );
        sqlx::query_as("SELECT * FROM doc_embedding WHERE domain = $1 AND path = $2 AND ver = $3")
            .bind(&doc.domain)
            .bind(&doc.path)
            .bind(use_ver)
            .fetch_optional(get_pg!())
            .await
            .unwrap()
    }

    pub async fn embed_document(doc: &Document, ver: &str, shell: &Shell) {
        if let Some(embed) = doc
            .embedding(Some(ver.to_string()), shell)
            .await
            .log_warn_none_and_pass(|| {
                format!(
                    "No embeds could be generated for {} / {}",
                    doc.domain, doc.path
                )
            })
        {
            let _ = sqlx::query(
                "DELETE FROM doc_embedding WHERE domain = $1 AND path = $2 AND ver = $3",
            )
            .bind(&doc.domain)
            .bind(&doc.path)
            .bind(ver)
            .execute(get_pg!())
            .await;

            for (index, embed) in embed.iter().enumerate() {
                sqlx::query("INSERT INTO doc_embedding VALUES ($1, $2, $3, $4, $5)")
                    .bind(&doc.domain)
                    .bind(&doc.path)
                    .bind(ver)
                    .bind(index as i64)
                    .bind(embed)
                    .execute(get_pg!())
                    .await
                    .unwrap();
            }
        }
    }

    pub async fn ensure_embedding(doc: &Document, shell: &Shell) {
        for ver in doc.versions() {
            if Self::get_embedding(doc, Some(ver.as_str())).await.is_none() {
                Self::embed_document(doc, &ver, shell).await;
            }
        }
    }

    pub async fn search_vector(v: &pgvector::Vector, limit: i64, offset: i64) -> Vec<SearchResult> {
        // limit should cover SearchResults not the query -> rework

        let results: Vec<DocEmbedding> = sqlx::query_as(
            "SELECT *, 1 / (1 + (embed_mxbai_embed_large <-> $1)) AS similarity FROM doc_embedding ORDER BY embed_mxbai_embed_large <=> $1 LIMIT $2 OFFSET $3",
        )
        .bind(v)
        .bind(limit)
        .bind(offset)
        .fetch_all(get_pg!())
        .await
        .unwrap();

        let mut search_res: HashMap<String, HashMap<String, SearchResult>> = HashMap::new();

        for res in results {
            let domain = search_res.entry(res.domain.clone()).or_default();
            let doc = domain.entry(res.path.clone()).or_insert(SearchResult::new(
                res.domain.clone(),
                res.path.clone(),
                res.total_chunks().await,
            ));
            doc.chunks.push(res);
        }

        let mut flat = search_res
            .into_values()
            .flat_map(|x| x.into_values().collect::<Vec<SearchResult>>())
            .collect::<Vec<SearchResult>>();

        flat.sort_by(|a, b| {
            b.similarity()
                .partial_cmp(&a.similarity())
                .unwrap_or(std::cmp::Ordering::Equal)
        });
        flat
    }

    pub async fn generate_embeddings_for(arc: &WebsiteArchive, shell: &Shell) {
        log::info!("Generating embeddings");

        for dom in arc.domains() {
            let dom = arc.get_domain(&dom);
            embed_path(&dom, "/", shell).await;
        }

        log::info!("Done generating embeddings");
    }
}

pub async fn embed_path(dom: &Domain, path: &str, shell: &Shell) {
    let (paths, is_doc) = dom.paths(path);

    // If the path is a document, process the root path.
    if is_doc {
        let doc = dom.path("/");
        EmbedStore::ensure_embedding(&doc, shell).await;
    }

    // Create a queue to process paths iteratively
    let mut queue = VecDeque::new();

    // Add the initial paths to the queue
    queue.extend(paths);

    while let Some(next_path) = queue.pop_front() {
        let (next_paths, is_doc) = dom.paths(next_path.path());

        if is_doc {
            let doc = dom.path(next_path.path());
            EmbedStore::ensure_embedding(&doc, shell).await;
        }

        queue.extend(next_paths);
    }
}
