use std::{io::Read, path::PathBuf};

use based::{request::RequestContext, ui::components::prelude::Shell};
use maud::html;

use crate::{blacklist::check_blacklist, render_page};

use super::read_dir;

/// Represents a document within a domain
pub struct Document {
    /// The domain associated with the document.
    pub domain: String,
    /// The path of the document within the domain.
    pub path: String,
    base_dir: PathBuf,
}

impl Document {
    /// Creates a new `Document` instance.
    ///
    /// # Parameters
    /// - `domain`: The domain to which the document belongs.
    /// - `path`: The path of the document within the domain.
    /// - `base_dir`: The base directory of the archive storage.
    ///
    /// # Returns
    /// A new `Document` instance.
    pub fn new(domain: &str, path: &str, base_dir: PathBuf) -> Self {
        let split = path
            .split('/')
            .filter(|x| !x.is_empty())
            .collect::<Vec<&str>>();

        Self {
            domain: domain.to_string(),
            path: if split.is_empty() {
                "/".to_string()
            } else {
                split.join("/")
            },
            base_dir,
        }
    }

    /// Renders the document, returning its content as a string.
    ///
    /// If the environment variable `$ROUTE_INTERNAL` is set to `true`, all links will be rewritten to point to internal archived routes.
    ///
    /// # Parameters
    /// - `version`: An optional version of the document to render in the format `YYYY-MM-DD`.
    ///
    /// # Returns
    /// An `Option` containing the rendered content as a string, or `None` if nothing could be rendered.
    pub async fn render_local(
        &self,
        version: Option<String>,
        shell: &Shell,
    ) -> Result<Vec<u8>, String> {
        if check_blacklist(&self.domain) {
            let content = html! {
                h3 { "This site is blacklisted" };
            };
            return Err(render_page(content, RequestContext::default(), shell)
                .await
                .1
                 .1);
        }

        let mut file_path = self.doc_dir();

        let latest_version = if let Some(version) = version {
            format!("index_{version}")
        } else {
            let versions = self.versions();
            let version = versions.first().cloned().ok_or(String::new())?;
            format!("index_{version}")
        };

        file_path = file_path.join(latest_version);

        let mut buf = Vec::new();
        std::fs::File::open(file_path)
            .map_err(|_| String::new())?
            .read_to_end(&mut buf)
            .unwrap();
        Ok(buf)
    }

    /// Determines the directory where the document is stored.
    ///
    /// # Returns
    /// A `PathBuf` representing the document directory.
    pub fn doc_dir(&self) -> PathBuf {
        let mut file_path = self.base_dir.join(&self.domain);

        for p in self.path.split('/').filter(|x| !x.is_empty()) {
            file_path = file_path.join(p);
        }

        file_path
    }

    fn latest_time_since(&self) -> Option<usize> {
        if let Some(t_str) = self.versions().first() {
            let given_date = chrono::NaiveDate::parse_from_str(t_str, "%Y-%m-%d")
                .expect("Invalid date format. Expected yyyy-mm-dd.");

            let today = chrono::Local::now().date_naive();

            let duration = today.signed_duration_since(given_date);
            return Some(duration.num_days() as usize);
        }

        None
    }

    /// Retrieves available versions of the document.
    ///
    /// # Returns
    /// A vector of strings representing the available versions of the document, sorted in descending order.
    pub fn versions(&self) -> Vec<String> {
        let mut res: Vec<String> = read_dir(&self.doc_dir())
            .into_iter()
            .filter_map(|x| {
                if x.starts_with("index_") {
                    return Some(x.trim_start_matches("index_").to_string());
                }

                None
            })
            .collect();
        res.sort();
        res.reverse();
        res
    }
}
