use std::path::PathBuf;

use based::result::LogAndIgnore;

use crate::blacklist::check_blacklist;

use super::{read_dir, Document};

/// Represents a domain within the website archive
pub struct Domain {
    /// Domain name
    pub name: String,
    dir: PathBuf,
}

impl Domain {
    /// Creates a new `Domain` instance.
    ///
    /// If the domain name is not blacklisted, a directory is created.
    ///
    /// # Parameters
    /// - `name`: The name of the domain.
    /// - `dir`: The directory path for the domain.
    ///
    /// # Returns
    /// A new `Domain` instance.
    pub fn new(name: &str, dir: PathBuf) -> Self {
        if !check_blacklist(name) {
            std::fs::create_dir_all(&dir)
                .log_err_and_ignore(&format!("Could not create domain dir {name}"));
        }
        Self {
            name: name.to_string(),
            dir,
        }
    }

    /// Resolves a specific path within the domain and returns a `Document` representing it.
    ///
    /// # Parameters
    /// - `path`: The path to resolve within the domain.
    ///
    /// # Returns
    /// A `Document` instance corresponding to the given path.
    pub fn path(&self, path: &str) -> Document {
        Document::new(&self.name, path, self.dir.parent().unwrap().to_path_buf())
    }

    /// Get all paths associated with the domain
    pub fn all_paths(&self) -> Vec<PathEntry> {
        let mut queue = self.paths("/").0;

        let mut ret = Vec::new();

        ret.push(PathEntry(self.name.clone(), "/".to_string()));

        while let Some(el) = queue.pop() {
            ret.push(el.clone());
            let paths = self.paths(&el.1).0;
            queue.extend(paths);
        }

        ret
    }

    /// Retrieves entries and metadata for a given path within the domain.
    ///
    /// # Parameters
    /// - `path`: The path to inspect.
    ///
    /// # Returns
    /// A tuple containing:
    /// - A vector of `PathEntry` instances representing the contents of the path.
    /// - A boolean indicating whether the path is itself a `Document`
    pub fn paths(&self, path: &str) -> (Vec<PathEntry>, bool) {
        let mut base_path = self.dir.clone();

        for p in path.split('/') {
            base_path = base_path.join(p);
        }

        let path = path
            .split("/")
            .filter(|x| !x.is_empty())
            .collect::<Vec<&str>>()
            .join("/");

        let dir_content = read_dir(&base_path);

        let mut ret = Vec::new();

        let mut is_doc = false;

        for entry in dir_content {
            let url_path = format!("{path}/{entry}");
            let url_path = url_path
                .split("/")
                .filter(|x| !x.is_empty())
                .collect::<Vec<&str>>()
                .join("/");
            if entry.starts_with("index_") {
                is_doc = true;
                continue;
            }

            ret.push(PathEntry(self.name.clone(), url_path));
        }

        (ret, is_doc)
    }
}

/// Represents an entry within a domain's path, containing its name and URL path.
#[derive(Debug, Clone)]
pub struct PathEntry(String, String);

impl PathEntry {
    pub fn url(&self) -> String {
        format!("/d/{}/{}", self.0, self.1)
    }

    pub fn path(&self) -> &String {
        &self.1
    }
}
