use std::sync::Arc;

use serde::Deserialize;
use tokio::sync::OnceCell;

pub static CONFIG: OnceCell<Arc<Config>> = OnceCell::const_new();

/// Get a reference to global config
pub fn get_config() -> &'static Arc<Config> {
    crate::conf::CONFIG.get().unwrap()
}

/// Load a global config
pub fn load_config(path: &str) {
    // TODO : Other load locations
    if let Ok(file_content) = std::fs::read_to_string(path) {
        let conf: Config =
            toml::from_str(&file_content).expect("Could not deserialize config file");
        crate::conf::CONFIG.set(std::sync::Arc::new(conf)).unwrap();
    }
}

/// Load a default global config
pub fn load_default_config() {
    if crate::conf::CONFIG.get().is_none() {
        crate::conf::CONFIG
            .set(std::sync::Arc::new(Config::default()))
            .unwrap();
    }
}

#[allow(non_snake_case)]
#[derive(Debug, Clone, Deserialize)]
pub struct Config {
    pub ROUTE_INTERNAL: bool,
    pub DOWNLOAD_ON_DEMAND: bool,
    pub ai: Option<AIConfig>,
    pub websites: Option<WebsiteConfig>,
}

impl Config {
    pub fn get_domain_config(&self, domain: &str) -> Option<&DomainConfig> {
        if let Some(websites) = &self.websites {
            if let Some(domains) = &websites.domains {
                let domain = domains.iter().find(|x| x.domain == domain);
                return domain;
            }
        }

        None
    }

    pub fn get_outdated(&self, domain: &str) -> Option<usize> {
        if let Some(conf) = self.get_domain_config(domain) {
            if let Some(outdated) = conf.outdated {
                return Some(outdated);
            }
        }

        if let Some(outdated) = self.websites.as_ref().map(|x| x.outdated) {
            return outdated;
        }

        None
    }
}

#[allow(non_snake_case)]
#[derive(Debug, Clone, Deserialize)]
pub struct AIConfig {
    pub OLLAMA_URL: String,
}

#[allow(non_snake_case)]
#[derive(Debug, Clone, Deserialize)]
pub struct WebsiteConfig {
    pub BLACKLIST_DOMAINS: Option<Vec<String>>,
    pub outdated: Option<usize>,
    pub keep_n: Option<usize>,
    pub domains: Option<Vec<DomainConfig>>,
}

#[derive(Debug, Clone, Deserialize)]
pub struct DomainConfig {
    pub domain: String,
    pub outdated: Option<usize>,
    pub keep_n: Option<usize>,
    pub blacklist_paths: Option<Vec<String>>,
    pub no_audio: Option<bool>,
    pub no_video: Option<bool>,
    pub no_image: Option<bool>,
    pub no_css: Option<bool>,
    pub no_javascript: Option<bool>,
    pub no_fonts: Option<bool>,
    pub no_frames: Option<bool>,
    pub user_agent: Option<String>,
}

impl Default for DomainConfig {
    fn default() -> Self {
        Self {
            domain: String::new(),
            blacklist_paths: None,
            outdated: None,
            keep_n: None,
            no_audio: Some(false),
            no_video: Some(false),
            no_image: Some(false),
            no_css: Some(false),
            no_javascript: Some(false),
            no_fonts: Some(false),
            no_frames: Some(false),
            user_agent: None,
        }
    }
}

impl Default for Config {
    fn default() -> Self {
        Self {
            ROUTE_INTERNAL: false,
            DOWNLOAD_ON_DEMAND: false,
            ai: None,
            websites: None,
        }
    }
}
