use std::vec;
use std::{collections::HashMap, io::Read, path::PathBuf, sync::Arc};

use based::ui::components::prelude::*;
use based::ui::prelude::*;
use based::ui::primitives::div::Center;
use based::{
    page,
    request::{
        api::GeneratedPager, assets::DataResponse, respond_html, respond_json, RequestContext,
        StringResponse,
    },
    ui::primitives::flex::Column,
};
use chrono::NaiveDate;
use itertools::Itertools;
use maud::{html, PreEscaped};
use regex::Regex;
use rocket::response::Redirect;
use rocket::{get, request::FromSegments, State};

pub mod component;
use component::*;
use serde_json::json;

use webarc::archive::{
    domain_has_fragments, get_domains_of_fragment, get_fragment, get_fragments_of_domain,
    get_random_fragment_id, internalize_urls, Document, DocumentIndex,
};
use webarc::get_mime_type;
use webarc::{
    ai::{generate_embedding, remove_data_urls, EmbedStore, SearchResult},
    archive::{extract_domains, WebsiteArchive},
    conf::get_config,
    render_page,
};

const SEARCH_BAR_STYLE: &str = "w-full px-4 mb-4 py-2 text-white bg-black border-2 border-neon-blue placeholder-neon-blue focus:ring-2 focus:ring-neon-pink focus:outline-none font-mono text-lg";

#[allow(non_snake_case)]
pub fn WebsiteIcon(domain: &str) -> PreEscaped<String> {
    html! {
        h2 class="text-xl font-bold mb-4 -ml-2 flex items-center w-fit" {
            img class="p-2 w-10 h-10" src=(format!("/favicon/{domain}")) {};
            a href=(format!("/d/{domain}")) { (domain) };
        };
    }
}

fn time_ago(naive_datetime: &chrono::NaiveDate) -> String {
    let now = chrono::Local::now().date_naive();
    let duration = now.signed_duration_since(*naive_datetime);

    if duration.num_seconds() < 60 {
        let s = duration.num_seconds();
        if s == 0 {
            "today".to_string()
        } else {
            format!("{} seconds ago", duration.num_seconds())
        }
    } else if duration.num_minutes() < 60 {
        format!("{} minutes ago", duration.num_minutes())
    } else if duration.num_hours() < 24 {
        format!("{} hours ago", duration.num_hours())
    } else if duration.num_days() < 30 {
        format!("{} days ago", duration.num_days())
    } else if duration.num_days() < 365 {
        format!("{} months ago", duration.num_days() / 30)
    } else {
        format!("{} years ago", duration.num_days() / 365)
    }
}

pub fn build_timeline(
    domains: HashMap<String, Vec<String>>,
    the_day: NaiveDate,
    vanish: bool,
) -> PreEscaped<String> {
    let mut tl = Timeline();

    let yesterday = the_day - chrono::Duration::days(1);

    let mut sorted_keys: Vec<_> = domains.keys().collect();
    sorted_keys.sort();

    for key in sorted_keys {
        let (domain, paths) = (key, domains.get(key).unwrap());
        tl = tl.add_element(TimelineElement::new(
            Tooltip(time_ago(&the_day), Text(&the_day.to_string())),
            WebsiteIcon(domain),
            Column(
                paths
                    .into_iter()
                    .map(|path| Link(&format!("/s/{domain}/{path}"), Text(&path).large()))
                    .collect(),
            )
            .gap(ScreenValue::_2),
        ));
    }

    tl = tl.add_after(InfinityScroll(
        ColoredSpinner(Purple::_500),
        &format!("/timeline?before={}", yesterday.to_string()),
    ));

    if vanish {
        tl = tl.vanish();
    }

    tl.render_with_class("")
}

pub async fn get_domains(before: &str, domain: Option<&str>) -> HashMap<String, Vec<String>> {
    let today = NaiveDate::parse_from_str(before, "%Y-%m-%d").unwrap();
    DocumentIndex::get_documents_of_day(today, domain).await
}

pub async fn get_domains_lookback(
    before: &str,
    mut days: u64,
    domain: Option<&str>,
) -> (HashMap<String, Vec<String>>, NaiveDate) {
    let mut the_day = NaiveDate::parse_from_str(before, "%Y-%m-%d").unwrap();
    while days > 0 {
        let domains = get_domains(&the_day.to_string(), domain).await;

        if !domains.is_empty() {
            return (domains, the_day);
        }

        the_day -= chrono::Duration::days(1);
        days -= 1;
    }

    (HashMap::new(), the_day)
}

#[get("/timeline?<before>&<domain>")]
pub async fn timeline_route(
    ctx: RequestContext,
    shell: &State<Shell>,
    before: Option<&str>,
    domain: Option<&str>,
) -> StringResponse {
    let today = chrono::Local::now().date_naive();

    let (domains, day) = if let Some(before) = before {
        get_domains_lookback(before, 30 * 12, domain).await
    } else {
        get_domains_lookback(&today.to_string(), 30 * 12, domain).await
    };

    if domains.is_empty() {
        return respond_html(Nothing().0);
    }

    if ctx.is_htmx && !ctx.htmx_redirect {
        return respond_html(build_timeline(domains, day, true).0);
    }

    let tl = Screen::medium(Padding(Nothing()).all(ScreenValue::_6))
        .on(Column(vec![
            Margin(Text("Document Timeline")._3xl().extrabold())
                .left(ScreenValue::_6)
                .render_with_class(""),
            build_timeline(domains, day, false),
        ]))
        .render_with_class("");
    page!(shell, ctx, "Timeline", tl)
}

/// Get the favicon of a domain
#[get("/favicon/<domain>")]
pub async fn favicon_route(domain: &str) -> Option<DataResponse> {
    let mut buf = Vec::new();
    std::fs::File::open(std::path::Path::new("./favicon").join(domain))
        .ok()?
        .read_to_end(&mut buf)
        .ok()?;

    // TODO : Default favicon

    Some(DataResponse::new(
        buf,
        "image/x-icon".to_string(),
        Some(60 * 60 * 24 * 5),
    ))
}

/// Websites Overview
#[get("/")]
pub async fn index(
    ctx: RequestContext,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let websites = arc.domains();

    let content = html! {
        div class="container mx-auto p-4" {

            div class="mb-4" {
                input type="search" name="query" placeholder="Search..." class=(SEARCH_BAR_STYLE)
                hx-get=("/vector_search")
                hx-target="#website_grid" hx-push-url="true" hx-swap="outerHTML" {};
            };


        div id="website_grid" {
            h1 class="text-5xl font-bold text-center mb-10" { "Websites" };
            div class="grid grid-cols-2 sm:grid-cols-3 lg:grid-cols-5 xl:grid-cols-6 2xl:grid-cols-8 gap-6" {
                @for site in websites {
                    a href=(format!("/d/{site}")) class="bg-neutral-900 shadow-md rounded-lg hover:bg-neutral-800 bg-gray-1 hover:cursor-pointer transition-all duration-300 flex flex-col items-center justify-center aspect-square max-w-60" {
                        div class="bg-blue-500 text-white rounded-full p-4" {
                            (favicon(&site))
                        };
                        p class="mt-4 text-base font-medium" { (site) };
                    };
                };
            };
        };
        }
    };

    render_page(content, ctx, &shell).await
}

/// Overview on <domain> / <path>
#[get("/d/<domain>/<paths..>")]
pub async fn domain_info_route(
    ctx: RequestContext,
    domain: &str,
    paths: PathBuf,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let domain = arc.get_domain(domain);
    let document = domain.path(paths.to_str().unwrap());
    let versions: Vec<String> = document.versions();
    let (path_entries, is_doc) = domain.paths(paths.to_str().unwrap());
    let path_seperations: Vec<&str> = paths.to_str().unwrap().split('/').collect();

    let domains = extract_domains(&String::from_utf8_lossy(
        &document
            .render_local(None, &shell)
            .await
            .unwrap_or_default(),
    ));

    // TODO : refactor

    let content = html! {
        h2 class="text-xl font-bold mb-4 flex items-center w-fit mx-auto" {
            img class="p-2" src=(format!("/favicon/{}", &domain.name)) {};
            a href=(format!("/d/{}", &domain.name)) { (domain.name) };
            (slash_seperator())
            (gen_path_header(path_seperations, &domain.name, true))
        };

        @if !versions.is_empty() {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md mb-8" {
                h3 class="font-bold mb-2" { "Saved Versions:" };
                @for (index, ver) in versions.into_iter().enumerate() {
                    a href=(format!("/s/{}/{}?time={}", &domain.name, paths.to_str().unwrap(), ver)) class="text-xs text-blue-400" {

                        @if index == 0 {
                        (format!("Latest ({ver})"))
                        } @else {
                            (ver)
                        }
                    };
                }
            };
        }

        @if path_entries.is_empty() && !is_doc {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
                p class="p-4 text-xl" { "Nothing found" };
            }
        }

        @if !path_entries.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Paths:" };
            ul class="space-y-2" {
                @for path in path_entries {
                    a href=(path.url()) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (arrow_icon("blue"))

                        span class="font-medium" { (path.path()) };
                    };
                };
            };
        };
    };

    @if !domains.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Domains linked on this page:" };
            ul class="space-y-2 p-4" {
                @for domain in domains {
                    a href=(format!("/d/{domain}")) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (favicon(&domain));
                        span class="font-medium" { (domain) };
                    };
                };
            };
        };
    };
    };

    render_page(content, ctx, &shell).await
}

#[get("/txt/<domain>/<path..>?<time>&<no_data_urls>")]
pub async fn render_txt_website(
    domain: &str,
    path: PathBuf,
    time: Option<&str>,
    no_data_urls: Option<&str>,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> Option<String> {
    let document = arc.get_domain(domain).path(path.to_str().unwrap());

    let content = document
        .render_local(time.map(|time| time.to_string()), &shell)
        .await
        .ok()?;

    let mut content = String::from_utf8_lossy(&content).to_string();

    if no_data_urls.is_some() {
        content = remove_data_urls(&content);
    }

    Some(html2md::parse_html(&content))
}

pub struct PathSegment {
    segments: Vec<String>,
}

impl std::fmt::Display for PathSegment {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.to_str())
    }
}

impl PathSegment {
    pub fn to_str(&self) -> String {
        self.segments.join("/")
    }
}

impl<'r> FromSegments<'r> for PathSegment {
    type Error = ();

    fn from_segments(
        segments: rocket::http::uri::Segments<'r, rocket::http::uri::fmt::Path>,
    ) -> Result<Self, Self::Error> {
        let paths: Vec<_> = segments
            .filter_map(|x| {
                if x == "." {
                    return None;
                }

                if x == ".." {
                    return None;
                }

                Some(x.to_string())
            })
            .collect();

        Ok(PathSegment { segments: paths })
    }
}

fn is_older_than(date_str: &str, num_days: usize) -> bool {
    let date =
        chrono::NaiveDate::parse_from_str(date_str, "%Y-%m-%d").expect("Invalid date format");
    let threshold_date = chrono::Utc::now().date_naive();
    let diff = threshold_date - date;
    diff.num_days() > num_days as i64
}

pub async fn redownload(
    arc: &WebsiteArchive,
    domain: &str,
    path: PathSegment,
    document: &Document,
    shell: &Shell,
) -> Option<DataResponse> {
    arc.archive_url(&format!("https://{domain}/{}", path.to_str()))
        .await;

    let mut content = document.render_local(None, &shell).await.ok()?;

    let mime = get_mime_type(&content).unwrap_or("text/html".to_string());

    if mime == "text/html" {
        if get_config().ROUTE_INTERNAL {
            content = internalize_urls(&String::from_utf8_lossy(&content), &domain)
                .as_bytes()
                .to_vec();
        }
    }

    return Some(DataResponse::new(content, mime, Some(60 * 60 * 24)));
}

/// Return archived version of `domain` / `path` at `time`
#[get("/s/<domain>/<path..>?<time>")]
pub async fn render_website(
    domain: &str,
    path: PathSegment,
    time: Option<&str>,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> Option<DataResponse> {
    let document = arc.get_domain(domain).path(&path.to_str());

    if time.is_none() {
        let versions = document.versions();
        if let Some(latest_version) = versions.first() {
            if let Some(outdated) = get_config().get_outdated(domain) {
                if is_older_than(latest_version, outdated) {
                    log::info!("Document {domain} / {path} is outdated, redownloading");
                    return redownload(&arc, domain, path, &document, &shell).await;
                }
            }
        }
    }

    let content = document
        .render_local(time.map(|time| time.to_string()), &shell)
        .await;

    // TODO : keep n versions

    if let Ok(mut content) = content {
        let mime = get_mime_type(&content).unwrap_or("text/html".to_string());

        if mime == "text/html" {
            // TODO : fragments url rewrite

            if get_config().ROUTE_INTERNAL {
                content = internalize_urls(&String::from_utf8_lossy(&content), &domain)
                    .as_bytes()
                    .to_vec();
            }

            // TODO : Fix
            // This will work if the `Content-Security-Policy` meta tag gets removed from the document
            // -> Migrate to HTML parser

            /*
            content = replace_data_urls(
                &String::from_utf8_lossy(&content),
                "http://127.0.0.1:8000"
            ).as_bytes().to_vec();
            */
        }

        return Some(DataResponse::new(content, mime, Some(60 * 60 * 24)));
    } else if get_config().DOWNLOAD_ON_DEMAND && time.is_none() {
        return redownload(&arc, domain, path, &document, &shell).await;
    }

    None
}

pub fn replace_data_urls(input: &str, root: &str) -> String {
    let data_url_pattern = r#"data:([a-zA-Z0-9]+/[a-zA-Z0-9.+-]+);base64,([a-zA-Z0-9+/=]+)"#;
    let re_data = Regex::new(data_url_pattern).unwrap();

    re_data
        .replace_all(input, |caps: &regex::Captures| {
            let encoded_data = caps.get(2).unwrap().as_str();

            if let Ok(decoded) = base64::decode(encoded_data) {
                let hash = webarc::sha256_hash(&decoded);
                format!("{root}/f/{hash}")
            } else {
                caps[0].to_string()
            }
        })
        .to_string()
}

pub fn gen_search_element(x: &SearchResult) -> PreEscaped<String> {
    html! {
        div class="text-xl font-bold mt-4 p-4 flex items-center w-full max-w-4xl max-h-40 mx-auto bg-neutral-800 shadow-md rounded-lg overflow-hidden border border-neutral-900 hover:cursor-pointer"
        hx-get=(format!("/d/{}/{}", x.domain, x.path))
        hx-target="#main_content" hx-push-url="true" hx-swap="innerHTML"
        {
                (favicon(&x.domain))
                a { (x.domain) };
                (slash_seperator());
                (gen_path_header(x.path.split('/').collect(), &x.domain, false));
                p class="font-bold p-2 text-stone-400" { (format!("{:.2} % [{} matches]", x.similarity() * 100.0, x.chunks.len())) };
        };
    }
}

#[get("/f/<hash>")]
pub async fn fragment_get(hash: &str) -> Option<DataResponse> {
    if hash == "random" {
        let random_hash = get_random_fragment_id().await;
        let (data, mime) = get_fragment(&random_hash).await?;
        return Some(DataResponse::new(data, mime, None));
    }

    let (data, mime) = get_fragment(hash).await?;
    Some(DataResponse::new(data, mime, Some(60 * 60 * 24)))
}

#[get("/vector_search?<query>&<page>")]
pub async fn vector_search(
    query: Option<&str>,
    page: Option<i64>,
    ctx: RequestContext,
    shell: &State<Shell>,
) -> Option<StringResponse> {
    get_config().ai.as_ref()?;

    let page = page.unwrap_or(1);

    // Search
    let search =
        Search::new("/vector_search".to_string()).search_class(SEARCH_BAR_STYLE.to_string());

    if let Some(query) = query {
        // If we have query
        let real_query = query.trim_end_matches(".json");

        // Search Results
        let vector = Arc::new(pgvector::Vector::from(
            generate_embedding(real_query.to_string()).await?,
        ));

        let results = GeneratedPager::new(
            |input: Arc<pgvector::Vector>, offset, limit| {
                Box::pin(async move {
                    EmbedStore::search_vector(&input, limit as i64, offset as i64).await
                })
            },
            1500,
        )
        .pager(page as u64, vector)
        .await;

        // API Route
        if query.ends_with(".json") {
            // TODO : Better search API
            return Some(respond_json(&json!(&results.page(page as u64))));
        }

        let content = search.build_response(&ctx, results, page, real_query, gen_search_element);

        return Some(render_page(content, ctx, &shell).await);
    }

    // Return new search site
    let content = search.build("", html! {});
    Some(render_page(content, ctx, &shell).await)
}

// TODO : better mime awereness in ui

#[allow(non_snake_case)]
pub fn MIMEIcon(mime: &str) -> PreEscaped<String> {
    if mime.starts_with("application") {
        return MaterialIcon("description");
    } else if mime.starts_with("audio") {
        return MaterialIcon("play_circle");
    } else if mime.starts_with("image") {
        return MaterialIcon("photo");
    } else if mime.starts_with("text") {
        return MaterialIcon("description");
    } else if mime.starts_with("video") {
        return MaterialIcon("movie");
    }

    MaterialIcon("description")
}

pub async fn mime_docs_of_domain(domain: &str) -> PreEscaped<String> {
    let other_mimes = DocumentIndex::get_documents_of_other_mime(domain).await;
    let by_mime = other_mimes
        .into_iter()
        .filter(|x| !x.mime.is_empty())
        .into_group_map_by(|x| x.mime.clone());

    println!("{domain} -> {by_mime:?}");

    if by_mime.keys().len() == 0 {
        return Nothing();
    }

    let first = by_mime.keys().next().unwrap();

    let mut t = Tabs().active(&format!("domain-{domain}-{first}"));

    for (mime, doc) in by_mime {
        t = t.add_tab(
            &format!("domain-{}-{mime}", domain),
            Row(vec![MIMEIcon(&mime), Text(&mime).render()])
                .items_center()
                .gap(ScreenValue::_2),
            UnorderedList().push_for_each(&doc, |doc: &_| {
                Link(
                    &doc.url(),
                    Context(
                        Hover(Background(Nothing()).color(Gray::_800)).on(Background(
                            Rounded(
                                Padding(
                                    Margin(
                                        Column(vec![
                                            Text(&doc.domain),
                                            Text(&doc.path),
                                            Text(&doc.version.to_string()),
                                        ])
                                        .gap(ScreenValue::_2),
                                    )
                                    .all(ScreenValue::_2),
                                )
                                .all(ScreenValue::_4),
                            )
                            .size(Size::Large),
                        )
                        .color(Gray::_700)),
                    ),
                )
            }),
        );
    }

    Column(vec![
        Margin(
            Row(vec![favicon(domain), Text(domain)._3xl().bold().render()]).gap(ScreenValue::_2),
        )
        .x(ScreenValue::_2)
        .top(ScreenValue::_4)
        .render(),
        t.render(),
    ])
    .gap(ScreenValue::_2)
    .render()
}

#[get("/mime")]
pub async fn mime_overview(
    ctx: RequestContext,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let domains = arc.domains();
    let mut d_html = Vec::new();

    for d in domains {
        d_html.push(mime_docs_of_domain(&d).await);
    }

    page!(
        shell,
        ctx,
        "Media Items",
        Margin(
            Div()
                .vanish()
                .push(Text("Media Items")._2xl().bold())
                .push_for_each(&d_html, |x: &_| x.render())
        )
        .all(ScreenValue::_4)
    )
}

#[get("/fragment/<fragment>")]
pub async fn fragment_overview(
    fragment: &str,
    ctx: RequestContext,
    shell: &State<Shell>,
) -> Option<StringResponse> {
    let fragment_info = get_fragment(fragment).await?;

    let docs = get_domains_of_fragment(fragment).await;

    let content = Margin(
        Column(vec![
            Center(Card(
                Row(vec![
                    Text(fragment).render(),
                    Rounded(
                        Padding(
                            Background(Text(&fragment_info.1).bold().black().xs())
                                .color(Colors::White),
                        )
                        .all(ScreenValue::_2),
                    )
                    .render(),
                    Link(&format!("/f/{fragment}"), Button("Raw")).render(),
                ])
                .gap(ScreenValue::_4)
                .items_center(),
            ))
            .render(),
            Center(Card(
                UnorderedList()
                    .push(Text("Referenced on:").bold()._2xl())
                    .push_for_each(&docs, |(domain, path, _): &_| {
                        Link(
                            &format!("/d/{domain}/{path}"),
                            Row(vec![
                                favicon(domain),
                                Text(domain).render(),
                                Text(path).render(),
                            ])
                            .items_center(),
                        )
                    }),
            ))
            .render(),
        ])
        .gap(ScreenValue::_4),
    )
    .top(ScreenValue::_4)
    .render();

    Some(page!(shell, ctx, "Fragment", content))
}

#[get("/fragments/<domain>")]
pub async fn fragments_domain_overview(
    domain: &str,
    ctx: RequestContext,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let fragments = get_fragments_of_domain(domain).await;

    let d_html: Vec<_> = fragments
        .into_iter()
        .map(|(fragment, mime)| {
            if mime.starts_with("image") {
                return Card(Link(
                    &format!("/fragment/{fragment}"),
                    Image(&format!("/f/{fragment}")).height(128).width(128),
                ));
            }

            Card(
                Tooltip(
                    Link(&format!("/fragment/{fragment}"), Text(&fragment)),
                    Text(&mime),
                )
                .white(),
            )
        })
        .collect();

    page!(
        shell,
        ctx,
        "Media Items",
        Margin(
            Div()
                .vanish()
                .push(
                    Margin(Text(&format!("Fragments of {domain}"))._2xl().bold())
                        .bottom(ScreenValue::_4)
                )
                .push(
                    Flex(Div().vanish().push_for_each(&d_html, |x: &_| x.render()))
                        .wrap(Wrap::Wrap)
                        .gap(ScreenValue::_3)
                        .justify(Justify::Around)
                )
        )
        .all(ScreenValue::_4)
    )
}

#[get("/fragments")]
pub async fn fragments_overview(
    ctx: RequestContext,
    arc: &State<WebsiteArchive>,
    shell: &State<Shell>,
) -> StringResponse {
    let domains = arc.domains();
    let mut d_html = Vec::new();

    for x in domains {
        if domain_has_fragments(&x).await {
            d_html.push(
                Link(
                    &format!("/fragments/{x}"),
                    Row(vec![favicon(&x), Text(&x).render()])
                        .items_center()
                        .render(),
                )
                .use_htmx()
                .render(),
            )
        }
    }

    page!(
        shell,
        ctx,
        "Media Items",
        Margin(
            Div()
                .vanish()
                .push(
                    Row(vec![
                        Margin(Text("Fragments")._2xl().bold())
                            .bottom(ScreenValue::_4)
                            .render(),
                        Margin(Button(
                            Row(vec![
                                MaterialIcon("casino"),
                                Link(&format!("/f/random"), "Random Fragment").render()
                            ])
                            .items_center()
                            .gap(ScreenValue::_2)
                        ))
                        .bottom(ScreenValue::_6)
                        .render()
                    ])
                    .gap(ScreenValue::_4)
                    .items_center()
                )
                .push_for_each(&d_html, |x: &_| x.render())
        )
        .all(ScreenValue::_4)
    )
}
